# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


from __future__ import absolute_import
import argparse
import time
import sys
from netq_lib.orm.redisdb import (
    set_connection_settings, reset_connection_settings,
    are_conn_and_listener_alive, BackendExceptions)
from netq_lib.common import utils
from netq_lib.common.service import Service
from netq_apps.lib import events
from netq_apps.lib import notifiers
from netq_apps.lib import applications
from netq_apps.lib.netq import NetQ
from netq_lib.common.enums import ConfKw
_PROC_NAME = 'netq-notifier'


class NotifierService(Service):

    def __init__(self, conf):
        super(NotifierService, self).__init__(conf, _PROC_NAME,
                                              enable_monkey_patch=False)
        self.pubsub_channel_dict = dict()
        self.events = dict()
        self.notifiers = dict()
        self.no_event_notifiers = []
        self.apps = dict()
        self.lib_netq = NetQ(self._conf[ConfKw.BACKEND], None, None)

        # Create App objects
        for app in applications.get_apps():
            obj = app(self._logger, conf)
            if obj.enable:
                self.apps[app.__name__] = obj

        # Create event objects and add them to pubsub channel dict
        for event in events.get_events():
            obj = event(self._logger)
            self.events[event.__name__] = obj
            self.pubsub_channel_dict[obj.channel_str] = obj.handler

        # Create Notifier Objects.
        # Update event objects with the right notifier handler to call
        # Update notifiers with the right app handlers
        for notifier in notifiers.get_notifiers():
            obj = notifier(self._logger, conf, self.lib_netq)
            if not obj.enable:
                continue
            self.notifiers[notifier.__name__] = obj
            if not obj.trigger_events:
                self.no_event_notifiers.append(obj)
            else:
                for event in obj.trigger_events:
                    event_obj = self.events[event]
                    event_obj.add_notifier_handler(obj.handler)
            for app, app_obj in self.apps.items():
                obj.add_app_handler(app_obj.handler)

    def run_notifiers(self):
        for _, notifier_obj in self.notifiers.items():
            notifier_obj.handler()

    def publish_msg(self, level, msg):
        for app, app_obj in self.apps.items():
            app_obj.handler([(level, 'NOTIFIER: %s' % msg)])

    def log_exception(self):
        self._logger.exception('NOTIFIER: %s' % sys.exc_info()[0])

    def _run(self):
        retry_time = 5
        conn_state = False
        self.publish_msg('info', 'Starting NetQ notification daemon')
        # Start listening to events
        try:
            set_connection_settings(self._conf[ConfKw.BACKEND],
                                    agent=False, subscribe=True,
                                    channels_dict=self.pubsub_channel_dict)
            conn_state = are_conn_and_listener_alive()
        except BackendExceptions:
            conn_state = False
        if not conn_state:
            self.publish_msg('warning',
                             ('Unable to connect to backend server. Retrying '
                              'every %d secs' % retry_time))
        else:
            try:
                self.run_notifiers()
            except BackendExceptions:
                conn_state = False
            except:
                self.log_exception()

        while True:
            time.sleep(retry_time)
            prev_conn_state = conn_state
            if not conn_state or not are_conn_and_listener_alive():
                try:
                    reset_connection_settings()
                    conn_state = are_conn_and_listener_alive()
                except BackendExceptions:
                    conn_state = False

            if prev_conn_state != conn_state:
                if prev_conn_state:
                    self.publish_msg(
                        'warning',
                        ('Lost connection to backend '
                         'server. Retrying every %d secs' % retry_time))
                    continue
                else:
                    self.publish_msg('info', 'Reconnected to backend server')
                    try:
                        self.run_notifiers()
                    except BackendExceptions:
                        conn_state = False
            try:
                if not conn_state:
                    continue
                self.lib_netq.refill_node_lst()
                for event, event_obj in self.events.items():
                    event_obj.call_notifier_handlers()
                for notifier in self.no_event_notifiers:
                    notifier.handler(None)
            except BackendExceptions:
                conn_state = False
            except:
                self.log_exception()


def main():
    parser = argparse.ArgumentParser(prog=_PROC_NAME,
                                     description='Netq Notifier Daemon')
    args = parser.parse_args()
    conf_dict = utils.get_yaml_config(_PROC_NAME)
    notifier_inst = NotifierService(conf_dict)
    return notifier_inst.run()
