# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
import socket
import time
import requests
import json
import abc
import inspect
import sys


def get_apps():
    app_lst = []
    apps = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for app in apps:
        if app[0] != 'Application':
            app_lst.append(app[1])
    return app_lst


class Application(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def populate_conf(self, main_conf):
        """ Read the config file and populate the class. Called by __init__ """

    @abc.abstractmethod
    def handler(self, msgs):
        """Called by notifiers to push messages to the application """

    def __init__(self, logger, main_conf):
        self.conf_name = self.__class__.__name__.lower()
        self._logger = logger
        self.enable = False
        self.header = None
        self.payload = None
        self.url = None
        self.populate_conf(main_conf)

    def get_enable(self, conf):
        enable = conf.get('enable', True)
        if isinstance(enable, str):
            if enable in ('no', 'false', 'off'):
                self.enable = False
        elif isinstance(enable, bool):
            self.enable = enable

    def app_warn_log(self, msg):
        self._logger.warn('%s: %s' % (self.conf_name.upper(), msg))


class Logger(Application):

    def populate_conf(self, main_conf):
        """ Logging is special and is controlled outside the application """
        self.enable = True

    def handler(self, msgs):
        if not msgs:
            return
        for msg in msgs:
            text = msg[1]
            if len(msg) > 2:
                text += '. They are: %s' % (', '.join(msg[2]))
            if msg[0].lower() == 'warning':
                self._logger.warning(text)
            else:
                self._logger.info(text)


class Slack(Application):
    webhook = None

    def populate_conf(self, main_conf):
        if not main_conf:
            return
        conf = main_conf.get(self.conf_name, None)
        if not conf:
            return
        self.get_enable(conf)
        self.webhook = conf.get('webhook', None)
        if self.enable and not self.webhook:
            self.app_warn_log('Notification disabled.'
                              '\'webhook\' missing in config file')
            self.enable = False

        self.header = {'content-type': 'application/json',
                       'Accept-Charset': 'UTF-8'}
        self.payload = {
            'username': 'netq notifier',
            'attachments': [{
             'footer': 'From netq-notifier running on %s '
                       % socket.gethostname(), }]}

    def handler(self, msgs):
        if not self.enable or not msgs:
            return
        for msg in msgs:
            kind = msg[0]
            title = None
            content = msg[1]
            text = None
            if kind.lower() == 'warning':
                title = ':warning:' + content + '\n'
                color = 'warning'
            elif kind.lower() == 'ok':
                title = ':+1:' + content + '\n'
                color = 'good'
            else:
                title = ':bulb:' + content + '\n'
                color = '#439FE0'
            if len(msg) > 2:
                if kind.lower() == 'ok':
                    text = ':shamrock:'
                    text += '\n:shamrock:'.join(sorted(msg[2]))
                else:
                    text = ':exclamation:'
                    text += '\n:exclamation:'.join(sorted(msg[2]))

            self.payload['attachments'][0]['ts'] = time.time()
            self.payload['attachments'][0]['title'] = title
            self.payload['attachments'][0]['color'] = color
            if text:
                self.payload['attachments'][0]['text'] = text
            else:
                if self.payload['attachments'][0].get('text'):
                    del self.payload['attachments'][0]['text']
            if self.webhook:
                r = requests.post(url=self.webhook,
                                  data=json.dumps(self.payload),
                                  headers=self.header)
                if r.status_code != requests.codes.ok:
                    self.app_warn_log('Request failed with code %d msg %s'
                                      % (r.status_code, r.text))


class PagerDuty(Application):
    api_access_key = None
    api_integration_key = None

    def populate_conf(self, main_conf):
        if not main_conf:
            return
        conf = main_conf.get(self.conf_name, None)
        if not conf:
            return
        self.get_enable(conf)
        self.api_access_key = conf.get('api_access_key', None)
        self.api_integration_key = conf.get('api_integration_key', None)
        if (self.enable and
                not self.api_access_key or not self.api_integration_key):
            self.app_warn_log(
                'Notifications disabled.\'API Access Key\' and/or \'Integration'
                ' key\' missing in config file')
            self.enable = False

        self.header = {
            'Accept': 'application/vnd.pagerduty+json;version=2',
            'Authorization': 'Token token=%s' % self.api_access_key,
            'Content-type': 'application/json'}
        self.url = 'https://events.pagerduty.com/v2/enqueue'

        self.payload = {
            'routing_key': self.api_integration_key,
            'event_action': 'trigger',
            'payload': {
                 'source': socket.gethostname(),
                 'component': 'netq-notifier'}
        }

    def handler(self, msgs):
        if not self.enable or not msgs:
            return
        for msg in msgs:
            kind = 'info'
            if msg[0].lower().startswith('warn'):
                kind = 'warning'
            if len(msg) > 2:
                self.payload['payload']['custom_details'] = sorted(msg[2])
            else:
                self.payload['payload']['custom_details'] = None
            self.payload['payload']['summary'] = msg[1]
            self.payload['payload']['severity'] = kind
            r = requests.post(url=self.url, data=json.dumps(self.payload),
                              headers=self.header)
            if r.status_code not in\
                    (requests.codes.ok, requests.codes.accepted):
                self._logger.warn('Request failed with code %d '
                                  'and msg %s' % (r.status_code, r.text))
