# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

from functools import wraps
import sys
import re

import ipaddr

from netq_lib.orm.query import Q
from netq_lib.orm import redisdb, restapi
from netq_apps.lib.netq import NetQ
from netq_lib.common.enums import (
    LinkType,
)

from netq_lib.orm.redisdb.models import (
    BgpSession,
    ClagSession,
    DockerContainer,
    DockerHost,
    DockerNetwork,
    DockerPortMap,
    DockerService,
    DockerSwarmNode,
    Link,
    Node,
    Services,
    OS,
    CPU,
    ASIC,
    Board,
    Disk,
    Memory,
    Fan,
    PSU,
    Temp,
    LnvSession,
    License
)


def get_hostname(args=None):

    hostname = '*'
    keywords = ['add', 'agent', 'check', 'example', 'help', 'notifier',
                'resolve', 'show', 'trace']

    if args:
        hostname = args.get('<hostname>')
        if not hostname:
            hostname = '*'
        elif isinstance(hostname, list):
            hostname = hostname[0].encode('ascii', 'ignore')
        else:
            hostname = hostname.encode('ascii', 'ignore')

    elif len(sys.argv) > 2:
        hostname = sys.argv[1]

        if hostname in keywords:
            hostname = '*'

    return hostname.encode('ascii', 'ignore')


def wrapMatcher(match_func):
    '''Decorator to setup redis server conn for match functions'''

    @wraps(match_func)
    def __wrapper(argv_word, use_fuzzy, tab_completing):
        keywords = ['around', 'changes', 'origin', 'json', 'vlan', 'bond',
                    'eth', 'swp', 'vlan', 'vxlan', 'loopback', 'vrf',
                    'macvlan', 'add', 'agent', 'check', 'example', 'help',
                    'notifier', 'resolve', 'show', 'trace']

        argv_word = argv_word.encode('ascii', 'ignore')
        if argv_word in keywords:
            return False

        if use_fuzzy:
            if ' ' not in argv_word:
                return True

        try:
            if ipaddr.IPAddress(argv_word):
                return False
        except ValueError:
            pass

        try:
            if ipaddr.IPNetwork(argv_word):
                return False
        except ValueError:
            if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                         argv_word) or
                re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                         argv_word)):
                return False

        server = NetQ.get_config_backend_first_server()
        server_addr = server.get('server') if server else None
        if not server_addr:
            out = [('<error>', 'Unable to contact remote DB')]
            return out
        if server_addr and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        elif server_addr:
            redisdb.set_connection_settings(NetQ.get_config_backend())

        choices = match_func(argv_word, use_fuzzy, tab_completing)

        if any(re.match(argv_word, choice) for choice in choices):
            return True

        if tab_completing:
            if any(choice.startswith(argv_word) for choice in choices):
                return True

        return False

    return __wrapper


def wrapCompleter(cmpl_func):
    '''Decorator to setup redis server conn for complete functions'''

    @wraps(cmpl_func)
    def __wrapper(args, ended_with_space, last_argv):

        server =  NetQ.get_config_backend_first_server()
        server_addr = server.get('server') if server else None
        if not server_addr:
            out = [('<error>', 'Unable to contact remote DB')]
            return out
        if server_addr and 'http' in server_addr:
            restapi.set_connection_settings(server_addr)
        elif server_addr:
            redisdb.set_connection_settings(NetQ.get_config_backend())

        return cmpl_func(args, ended_with_space, last_argv)

    return __wrapper


@wrapCompleter
def hostname_complete(args, ended_with_space, last_argv):
    """
    Completion for hostname
    """
    keywords = {
        'add': 'Add IP address or hostname of netq server',
        'agent': 'Start/Stop/Restart/Query status of netq agent',
        'check': 'Perform fabric-wide checks',
        'example': 'Show examples of usage and workflow',
        'help': 'Show usage info',
        'notifier': 'Start/Stop/Restart/Query status of netq notifier',
        'resolve': 'Annotate input with names and interesting info',
        'show': 'Show info about specified object',
        'trace': 'Control plane trace path across fabric'
    }

    out = []

    if last_argv:
        for node in Node.query.all():
            if node.hostname.startswith(last_argv):
                out.append((node.hostname, 'Host ' + node.hostname))

        for word, val in keywords.iteritems():
            if word.startswith(last_argv):
                out.append((word, val))

    if not out:
        out.append(('<hostname>',
                    'Type first char of netq host for dynamic completion'))

    return out


@wrapMatcher
def hostname_match(argv_word, use_fuzzy, tab_completing=False):
    """hostname matcher"""
    return [node.hostname for node in Node.query.filter(
        hostname=argv_word + '*')]


@wrapCompleter
def interface_complete(args, ended_with_space, last_argv):
    """
    Completion for remote interface name
    """
    out = []
    hostname = get_hostname(args)

    plinks = Link.query.filter(hostname=hostname,
                               exclude=Q(ifname='.*-v0'))
    for link in plinks:
        if ended_with_space or link.ifname.startswith(last_argv):
            out.append((link.ifname, link.ifname))

    if not out:
        out.append(('<remote-interface>', 'Interface name on %s' %
                    hostname))
    return out


@wrapCompleter
def physical_interface_complete(args, ended_with_space, last_argv):
    """
    Completion for remote physical interface name
    """
    out = []
    hostname = get_hostname(args)

    links = Link.query.filter(hostname=hostname,
                              exclude=Q(ifname='.*-v0'))
    for link in links:
        if link.kind in (LinkType.BOND, LinkType.SWP):
            if ended_with_space or link.ifname.startswith(last_argv):
                out.append((link.ifname, link.ifname))

    if not out:
        out.append(('<remote-physical-interface>', 'Interface name on %s' %
                    hostname))
    return out


@wrapMatcher
def interface_match(argv_word, use_fuzzy, tab_completing=False):
    """
    Match for remote interface name
    """

    hostname = get_hostname()
    return sorted(link.ifname for link in Link.query.filter(
        hostname=hostname, ifname=argv_word + '*'))


@wrapMatcher
def mode_match(argv_word, use_fuzzy, tab_completing=False):
    """
    Match show docker swarm service [<hostname>] [<service-name> | <mode>]
                                    [around <text-time>] [json]
    """

    hostname = get_hostname()
    return ([service.mode
             for service in DockerService.query.filter(hostname=hostname)])


@wrapCompleter
def mode_complete(args, ended_with_space, last_argv):
    """
    Completion for mode in
    show docker swarm service <hostname> [<service-name> | <mode>] ...
    """
    out = []

    hostname = get_hostname(args)
    services = DockerService.query.filter(hostname=hostname)

    for service in services:
        if ended_with_space or service.mode.startswith(last_argv):
            out.append((service.mode, '%s mode' % service.mode))

    if not out:
        out.append(('<mode>', 'Docker Swarm Node Mode'))
    return out


@wrapMatcher
def swarm_service_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                        tab_completing=False):  # pylint: disable=unused-argument
    """
    Match show docker swarm service [<hostname>] [<service-name> | <mode>]
                                    [around <text-time>] [json]
    """
    return ([service.service_name
             for service in DockerService.query.all()])


@wrapCompleter
def swarm_service_complete(args, ended_with_space, last_argv):
    """
    Completion for mode in
    show docker swarm service <hostname> [<service-name> | <mode>] ...
    """
    out = []

    services = DockerService.query.all()

    for service in services:
        if ended_with_space or service.service_name.startswith(last_argv):
            out.append((service.service_name,
                        'Swarm Service %s' % service.service_name))

    if not out:
        out.append(('<swarm-service-name>', 'Docker Swarm Service Name'))
    return out


@wrapMatcher
def network_name_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                       tab_completing=False):  # pylint: disable=unused-argument
    """
    Match network-name
    """
    hostname = get_hostname()
    return ([network.network_name
             for network in DockerNetwork.query.filter(hostname=hostname)])


@wrapCompleter
def network_name_complete(args, ended_with_space, last_argv):
    """
    Completion for <network-name> in
    show docker network [<hostname>] [<network-name> | <ipv4/prefixlen> ]
                                    [around <text-time>] [json]
    """
    out = []

    hostname = get_hostname(args)
    networks = DockerNetwork.query.filter(hostname=hostname)

    for network in networks:
        if ended_with_space or network.network_name.startswith(last_argv):
            out.append((network.network_name, 'Docker Network Name'))

    if not out:
        out.append(('<network-name>', 'Docker Network Name'))
    return out


@wrapCompleter
def network_driver_complete(args, ended_with_space, last_argv):
    """
    Completion for <network-type> in
    show docker network ...
    """
    out = []
    hostname = get_hostname(args)

    networks = DockerNetwork.query.filter(hostname=hostname)

    for network in networks:
        if ended_with_space or network.driver.startswith(last_argv):
            out.append((network.driver, 'Docker Network Type'))

    if not out:
        out.append(('<network-type>', 'Docker Network type'))
    return out


@wrapMatcher
def network_driver_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                         tab_completing=False):  # pylint: disable=unused-argument
    """
    Match network-type
    """
    hostname = get_hostname()
    return [network.driver
            for network in DockerNetwork.query.filter(hostname=hostname)]


@wrapMatcher
def proto_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                tab_completing=False):  # pylint: disable=unused-argument
    """
    Match network-name
    """

    hostname = get_hostname()
    return [portmap.proto
            for portmap in DockerPortMap.query.filter(hostname=hostname)]


@wrapCompleter
def proto_complete(args, ended_with_space, last_argv):
    """
    Completion for <proto> in
    show docker [<hostname>] container portmap [protocol <proto> [<port>]]
    """
    out = []

    hostname = get_hostname(args)

    portmaps = DockerPortMap.query.filter(hostname=hostname)
    for portmap in portmaps:
        if ended_with_space or portmap.proto.startswith(last_argv):
            out.append((portmap.proto, 'Protocol'))

    if not out:
        out.append(('<proto>', 'IP Protocol'))
    return out


def port_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
               tab_completing=False):  # pylint: disable=unused-argument
    """
    Match network-name
    """
    keywords = ['around', 'json']

    if argv_word in keywords:
        return False

    if (re.match(r'(^\d+$)', argv_word)):
        return True
    return False


@wrapMatcher
def cluster_id_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                     tab_completing=False):  # pylint: disable=unused-argument
    """
     Match for <cluster-name> in
     show docker [<hostname>] swarm cluster [<cluster-name> | <cluster-id>]
     show docker [<hostname>] swarm node  [<cluster-name> | <cluster-id>] [<role>]
    """
    hostname = get_hostname()

    return [node.cluster_id
            for node in DockerSwarmNode.query.filter(hostname=hostname)]


@wrapCompleter
def cluster_id_complete(args, ended_with_space, last_argv):
    """
     Match for <cluster-id> in
     show docker [<hostname>] swarm cluster [<cluster-name> | <cluster-id>]
     show docker [<hostname>] swarm node  [<cluster-name> | <cluster-id>] [<role>]
    """
    out = []

    hostname = get_hostname(args)
    nodes = DockerSwarmNode.query.filter(hostname=hostname)
    for node in nodes:
        if ended_with_space or node.cluster_id.startswith(last_argv):
            out.append((node.cluster_id, 'Cluster-id'))

    if not out:
        out.append(('<cluster-id>', 'Cluster ID'))
    return out


@wrapMatcher
def cluster_name_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                       tab_completing=False):  # pylint: disable=unused-argument
    """
     Match for <cluster-name> in
     show docker [<hostname>] swarm cluster [<cluster-name> | <cluster-id>]
     show docker [<hostname>] swarm node  [<cluster-name> | <cluster-id>] [<role>]
    """

    hostname = get_hostname()
    return [node.cluster_name
            for node in DockerSwarmNode.query.all(hostname=hostname)]


@wrapCompleter
def cluster_name_complete(args, ended_with_space, last_argv):
    """
    Completion for <cluster-name> in
     show docker [<hostname>] swarm cluster [<cluster-name> | <cluster-id>]
     show docker [<hostname>] swarm node  [<cluster-name> | <cluster-id>] [<role>]
    """
    out = []
    hostname = get_hostname(args)

    nodes = DockerSwarmNode.query.filter(hostname=hostname)
    for node in nodes:
        if ended_with_space or node.cluster_name.startswith(last_argv):
            out.append((node.cluster_name, 'Cluster Name'))

    if not out:
        out.append(('<cluster-name>', 'Cluster Name'))
    return out


@wrapMatcher
def role_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
               tab_completing=False):  # pylint: disable=unused-argument
    """
    Match network-name
    """

    hostname = get_hostname()
    return [node.role
            for node in DockerSwarmNode.query.filter(hostname=hostname)]


@wrapCompleter
def role_complete(args, ended_with_space, last_argv):
    """
    Completion for <role> in
    show docker [<hostname>] swarm node  [<cluster-id>] [<role>]
    """
    out = []

    hostname = get_hostname(args)
    nodes = DockerSwarmNode.query.filter(hostname=hostname)
    for node in nodes:
        if ended_with_space or node.role.startswith(last_argv):
            out.append((node.role, 'Role'))

    if not out:
        out.append(('<role>', 'Role'))
    return out


@wrapMatcher
def container_name_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                         tab_completing=False):  # pylint: disable=unused-argument
    """
    Match container-name
    """

    hostname = get_hostname()
    return [container.name
            for container in DockerContainer.query.filter(hostname=hostname)]


@wrapCompleter
def container_name_complete(args, ended_with_space, last_argv):
    """
    Completion for <container-name> in
    """
    out = []

    hostname = get_hostname(args)

    containers = DockerContainer.query.filter(hostname=hostname)
    for container in containers:
        if ended_with_space or container.name.startswith(last_argv):
            out.append((container.name, 'Container Name'))

    if not out:
        out.append(('<name>', 'Container Name'))
    return out


@wrapMatcher
def container_id_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                       tab_completing=False):  # pylint: disable=unused-argument
    """
    Match container-id
    """

    hostname = get_hostname()
    return ([container.container_id
             for container in DockerContainer.query.filter(hostname=hostname)])


@wrapCompleter
def container_id_complete(args, ended_with_space, last_argv):
    """
    Completion for <container-id> in
    show docker [<hostname>] container detail [<ipv4> | <ipv6> | <name> | <image> | <container-id>]
    """
    out = []

    hostname = get_hostname(args)

    containers = DockerContainer.query.filter(hostname=hostname)
    for container in containers:
        if ended_with_space or container.container_id.startswith(last_argv):
            out.append((container.container_id, 'Container Id'))

    if not out:
        out.append(('<container-id>', 'Container Id'))
    return out


@wrapMatcher
def container_image_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                          tab_completing=False):  # pylint: disable=unused-argument
    """
    Match container-image
    """
    hostname = get_hostname()

    return [container.image
            for container in DockerContainer.query.filter(hostname=hostname)]


@wrapCompleter
def container_image_complete(args, ended_with_space, last_argv):
    """
    Completion for <container-image>
    """
    out = []

    hostname = get_hostname(args)
    containers = DockerContainer.query.filter(hostname=hostname)
    for container in containers:
        if ended_with_space or container.image.startswith(last_argv):
            out.append((container.image, 'Container Image'))

    if not out:
        out.append(('<image>', 'Container Image'))
    return out


@wrapMatcher
def container_mac_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                        tab_completing=False):  # pylint: disable=unused-argument
    """
    Match <container-mac>
    """

    hostname = get_hostname()
    return [container.mac
            for container in DockerContainer.query.filter(hostname=hostname)]


@wrapCompleter
def container_mac_complete(args, ended_with_space, last_argv):
    """
    Completion for <container-mac> in
    """
    out = []

    hostname = get_hostname(args)

    containers = DockerContainer.query.filter(hostname=hostname)
    for container in containers:
        if ended_with_space or container.mac.startswith(last_argv):
            out.append((container.mac, 'Container MAC'))

    if not out:
        out.append(('<container-mac>', 'Container MAC'))
    return out


@wrapCompleter
def docker_version_complete(args, ended_with_space, last_argv):
    """
    Completion for <docker-version>
    """
    out = []

    hostname = get_hostname(args)
    hosts = DockerHost.query.filter(hostname=hostname)
    for host in hosts:
        if ended_with_space or host.docker_version.startswith(last_argv):
            out.append((host.docker_version, 'Docker Version'))

    if not out:
        out.append(('<docker-version>', 'Docker Version'))
    return out


@wrapMatcher
def docker_version_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                         tab_completing=False):  # pylint: disable=unused-argument
    """
    Match <docker-version>
    """
    hostname = get_hostname()
    return [host.docker_version
            for host in DockerHost.query.filter(hostname=hostname)]


def generic_match(argv_word, use_fuzzy,  # pylint: disable=unused-argument
                  tab_completing=False):  # pylint: disable=unused-argument
    """
    Match for os entries
    """
    keywords = ['around', 'changes', 'origin', 'json', 'vlan', 'bond', 'temp',
                'bridge', 'eth', 'swp', 'vlan', 'vxlan', 'loopback', 'fan',
                'psu', 'all', 'brief', 'macvlan']

    if argv_word in keywords:
        return False

    try:
        if ipaddr.IPAddress(argv_word):
            return False
    except ValueError:
        pass

    try:
        if ipaddr.IPNetwork(argv_word):
            return False
    except ValueError:
        if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                     argv_word) or
            re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                     argv_word)):
            return False

    return True


@wrapCompleter
def os_version_complete(args, ended_with_space, last_argv):
    """
    Completion for os version
    """
    out = []
    hostname = get_hostname(args)
    oses = OS.query.filter(hostname=hostname)

    for oss in oses:
        if ended_with_space or oss.version_id.startswith(last_argv):
            out.append((oss.version_id,
                        '%s version' % oss.name))
    if not out:
        out.append(('<os-version>', 'Operating System version'))
    return out


@wrapCompleter
def os_name_complete(args, ended_with_space, last_argv):
    """
    Completion for os version
    """
    out = []

    hostname = get_hostname(args)
    oses = OS.query.filter(hostname=hostname)

    for os_entry in oses:
        if ended_with_space or (last_argv.lower() in os_entry.name.lower()):
            out.append((os_entry.name, 'Operating System name'))
    if not out:
        out.append(('<os-name>', 'Operating System name'))
    return out


@wrapCompleter
def cpu_arch_complete(args, ended_with_space, last_argv):
    """
    Completion for processor architecture
    """
    out = []

    hostname = get_hostname(args)
    procs = CPU.query.filter(hostname=hostname)

    for proc in procs:
        if ended_with_space or (last_argv.lower() in proc.arch.lower()):
            out.append((proc.arch, 'CPU Architecture'))
    if not out:
        out.append(('<cpu-arch>', 'CPU Architecture'))
    return out


@wrapCompleter
def cpu_model_complete(args, ended_with_space, last_argv):
    """
    Completion for processor model
    """
    out = []
    hostname = get_hostname(args)
    procs = CPU.query.filter(hostname=hostname)

    for proc in procs:
        if ended_with_space or (last_argv.lower() in proc.model.lower()):
            out.append((proc.model, 'CPU Model'))
    if not out:
        out.append(('<cpu-model>', 'CPU Model'))
    return out


@wrapCompleter
def asic_vendor_complete(args, ended_with_space, last_argv):
    """
    Completion for network ASIC vendor
    """
    out = []
    hostname = get_hostname(args)
    procs = ASIC.query.filter(hostname=hostname)

    for proc in procs:
        if ended_with_space or (last_argv.lower() in proc.vendor.lower()):
            out.append((proc.vendor, 'Network ASIC Vendor'))
    if not out:
        out.append(('<asic-vendor>', 'Network ASIC Vendor'))
    return out


@wrapCompleter
def asic_model_complete(args, ended_with_space, last_argv):
    """
    Completion for network processor model
    """
    out = []
    hostname = get_hostname(args)
    procs = ASIC.query.filter(hostname=hostname)

    for proc in procs:
        if ended_with_space or (last_argv.lower() in proc.model.lower()):
            out.append((proc.model, 'Network ASIC Model'))
    if not out:
        out.append(('<asic-model>', 'Network ASIC Model'))
    return out


@wrapCompleter
def asic_model_id_complete(args, ended_with_space, last_argv):
    """
    Completion for network processor model id
    """
    out = []

    hostname = get_hostname(args)
    procs = ASIC.query.filter(hostname=hostname)

    for proc in procs:
        if ended_with_space or (last_argv.lower() in proc.model_id.lower()):
            out.append((proc.model_id, 'Network ASIC Model ID'))
    if not out:
        out.append(('<asic-model-id>', 'Network ASIC Model ID'))
    return out


@wrapCompleter
def board_vendor_complete(args, ended_with_space, last_argv):
    """
    Completion for board vendor
    """
    out = []
    hostname = get_hostname(args)
    boards = Board.query.filter(hostname=hostname)

    for board in boards:
        if ended_with_space or (last_argv.lower() in board.vendor.lower()):
            out.append((board.vendor, 'Board Vendor'))

    if not out:
        out.append(('<board-vendor>', 'Board Vendor'))
    return out


@wrapCompleter
def board_model_complete(args, ended_with_space, last_argv):
    """
    Completion for switch model
    """
    out = []
    hostname = get_hostname(args)
    boards = Board.query.filter(hostname=hostname)

    for board in boards:
        if ended_with_space or (last_argv.lower() in board.model.lower()):
            out.append((board.model, 'Board Model'))

    if not out:
        out.append(('<board-model>', 'Board Model'))
    return out


@wrapCompleter
def disk_name_complete(args, ended_with_space, last_argv):
    """
    Completion for disk name
    """
    out = []

    hostname = get_hostname(args)
    disks = Disk.query.filter(hostname=hostname)

    for disk in disks:
        if ended_with_space or (last_argv.lower() in disk.name.lower()):
            out.append((disk.name, 'Disk Name'))

    if not out:
        out.append(('<disk-name>', 'Disk Name'))
    return out


@wrapCompleter
def disk_size_complete(args, ended_with_space, last_argv):
    """
    Completion for disk size
    """
    out = []

    hostname = get_hostname(args)
    disks = Disk.query.filter(hostname=hostname)

    for disk in disks:
        if ended_with_space or (disk.lower().startswith(last_argv.lower())):
            out.append((disk.size, 'Disk Size'))

    if not out:
        out.append(('<disk-size>', 'Disk Size'))
    return out


@wrapCompleter
def disk_transport_complete(args, ended_with_space, last_argv):
    """
    Completion for disk transport
    """
    out = []
    hostname = get_hostname(args)
    disks = Disk.query.filter(hostname=hostname)

    for disk in disks:
        if ended_with_space or (last_argv.lower() in disk.transport.lower()):
            out.append((disk.size, 'Disk Transport'))

    if not out:
        out.append(('<disk-transport>', 'Disk Transport'))
    return out


@wrapCompleter
def disk_vendor_complete(args, ended_with_space, last_argv):
    """
    Completion for disk vendor
    """
    out = []
    hostname = get_hostname(args)
    disks = Disk.query.filter(hostname=hostname)

    for disk in disks:
        if ended_with_space or (last_argv.lower() in disk.vendor.lower()):
            out.append((disk.vendor, 'Disk Vendor'))

    if not out:
        out.append(('<disk-vendor>', 'Disk Vendor'))
    return out


@wrapCompleter
def memory_type_complete(args, ended_with_space, last_argv):
    """
    Completion for memory type
    """
    out = []
    hostname = get_hostname(args)
    memorys = Memory.query.filter(hostname=hostname)

    for memory in memorys:
        if ended_with_space or (last_argv.lower() in memory.m_type.lower()):
            out.append((memory.m_type, 'Memory type'))

    if not out:
        out.append(('<memory-type>', 'Memory type'))
    return out


@wrapCompleter
def memory_size_complete(args, ended_with_space, last_argv):
    """
    Completion for memory size
    """
    out = []
    hostname = get_hostname(args)
    memorys = Memory.query.filter(hostname=hostname)

    for memory in memorys:
        if (ended_with_space or (memory.size.lower().startswith(
                last_argv.lower()))):
            out.append((memory.size, 'Memory size'))

    if not out:
        out.append(('<memory-size>', 'Memory Size'))
    return out


@wrapCompleter
def memory_vendor_complete(args, ended_with_space, last_argv):
    """
    Completion for memory vendor
    """
    out = []
    hostname = get_hostname(args)
    memorys = Memory.query.filter(hostname=hostname)

    for memory in memorys:
        if ended_with_space or (last_argv.lower() in memory.vendor.lower()):
            out.append((memory.vendor, 'Memory vendor'))

    if not out:
        out.append(('<memory-vendor>', 'Memory Vendor'))
    return out


@wrapCompleter
def psu_name_complete(args, ended_with_space, last_argv):
    """
    Completion for psu names
    """
    out = []

    hostname = get_hostname(args)
    psus = PSU.query.filter(hostname=hostname)

    for psu in psus:
        if ended_with_space or (last_argv.lower() in psu.s_name.lower()):
            out.append((psu.s_name, 'Power Supply'))

    if not out:
        out.append(('<psu-name>', 'Power Supply Name'))
    return out


@wrapCompleter
def fan_name_complete(args, ended_with_space, last_argv):
    """
    Completion for fan names
    """
    out = []
    hostname = get_hostname(args)
    fans = Fan.query.filter(hostname=hostname)

    for fan in fans:
        if ended_with_space or (last_argv.lower() in fan.s_name.lower()):
            out.append((fan.s_name, 'Fan Name'))

    if not out:
        out.append(('<fan-name>', 'Fan Name'))
    return out


@wrapCompleter
def temp_name_complete(args, ended_with_space, last_argv):
    """
    Completion for Temp names
    """
    out = []
    hostname = get_hostname(args)
    temps = Temp.query.filter(hostname=hostname)

    for temp in temps:
        if ended_with_space or (last_argv.lower() in temp.s_name.lower()):
            out.append((temp.s_name, 'Temp Name'))

    if not out:
        out.append(('<temp-name>', 'Temp Name'))
    return out


def ip_server_match(argv_word,
                    use_fuzzy,  # pylint: disable=unused-argument
                    tab_completing=False):  # pylint: disable=unused-argument
    """
    Match only for IP addresses
    """
    try:
        if ipaddr.IPAddress(argv_word):
            return True
    except ValueError:
        pass
    return False


@wrapCompleter
def vrf_complete(args, ended_with_space, last_argv):
    """
    Completion for list of VRFs
    """
    out = []

    if ended_with_space or 'default'.startswith(last_argv):
        out.append(('default', 'default'))
    for link in Link.query.filter(kind='vrf'):
        if ended_with_space or link.ifname.startswith(last_argv):
            out.append((link.ifname, link.ifname))

    if not out:
        out.append(('<vrf>', 'VRF name'))
    return out


@wrapMatcher
def vrf_match(argv_word,
              use_fuzzy,  # pylint: disable=unused-argument
              tab_completing=False):  # pylint: disable=unused-argument
    """
    """
    hostname = get_hostname()

    choices = [link.ifname
               for link in Link.query.filter(hostname=hostname, kind='vrf')]
    choices.append('default')
    return choices


@wrapCompleter
def service_complete(args, ended_with_space, last_argv):
    """
    Completion for list of VRFs
    """
    out = []
    for entry in Services.query.filter(name='*'):
        if ended_with_space or entry.name.startswith(last_argv):
            out.append((entry.name, entry.name))

    if not out:
        out.append(('<service-name>', 'Name of service'))

    return out


@wrapMatcher
def service_match(argv_word,
                  use_fuzzy,
                  tab_completing=False):

    hostname = get_hostname()
    return [svc.name
            for svc in Services.query.filter(hostname=hostname, name='*')]


def server_name_match(argv_word,
                      use_fuzzy,  # pylint: disable=unused-argument
                      tab_completing=False):  # pylint: disable=unused-argument
    keywords = ['around', 'changes', 'origin', 'json', 'vlan', 'bond',
                'bridge', 'eth', 'swp', 'vlan', 'vrf', 'vxlan', 'loopback',
                'macvlan', 'psu', 'fan', 'temp']

    if argv_word in keywords:
        return False

    try:
        if ipaddr.IPAddress(argv_word):
            return False
    except ValueError:
        pass

    try:
        if ipaddr.IPNetwork(argv_word):
            return False
    except ValueError:
        if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                     argv_word) or
                re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                         argv_word)):
            return False

    return True


def inventory_match(argv_word,
                    use_fuzzy,  # pylint: disable=unused-argument
                    tab_completing=False):  # pylint: disable=unused-argument
    keywords = ['around', 'changes', 'origin', 'json', 'vlan', 'bond',
                'bridge', 'eth', 'swp', 'vlan', 'vrf', 'vxlan', 'loopback',
                'macvlan']

    if argv_word in keywords:
        return False

    try:
        if ipaddr.IPAddress(argv_word):
            return False
    except ValueError:
        pass

    try:
        if ipaddr.IPNetwork(argv_word):
            return False
    except ValueError:
        if (re.match(r'(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}',
                     argv_word) or
                re.match(r'[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}',
                         argv_word)):
            return False

    if argv_word in ['os', 'board', 'disk', 'cpu', 'asic', 'memory']:
        return True

    return True
