# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

from netq_lib.orm.redisdb import models
import abc
import inspect
import sys

'''
To add a new notifier, all you need to specify are the "trigger_events"
 (see PubSubEvents in events.py) and then "run(self, events_map)"
 E.g. For Vlan
 class Vlan(Notifier):
    trigger_events = ['Link']

    def run(self, event_map):
        self.mismatch_run(self.lib_netq.check_vlan_consistency)

 Config in conf file is same for all notifiers.
 <class_name>:
   enable: true
   exclude: []
   include: []
'''


def get_notifiers():
    notifier_lst = []
    notifiers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for notifier in notifiers:
        if notifier[0] != 'Notifier':
            notifier_lst.append(notifier[1])
    return notifier_lst


class Notifier(object):
    __metaclass__ = abc.ABCMeta
    trigger_events = None

    def __init__(self, logger, main_conf, lib_netq):
        self.conf_name = self.__class__.__name__.lower()
        self._logger = logger
        self.lib_netq = lib_netq
        self.msgs = []
        self.app_handlers = []
        self.enable = True
        self.include = set()
        self.exclude = set()
        self.log_str = self.conf_name.upper()
        self.failed_set = None
        self.populate_conf(main_conf)

    @abc.abstractmethod
    def run(self, events_map):
        """ Run the check and append messages as a tuple of
            ('log', 'msg', 'list')  in self.msgs """

    def populate_conf(self, main_conf):
        if not main_conf:
            return
        conf = main_conf.get(self.conf_name, None)
        if not conf:
            return
        enable = conf.get('enable', True)
        if isinstance(enable, str):
            if enable in ('no', 'false', 'off'):
                self.enable = False
        elif isinstance(enable, bool):
            self.enable = enable
        self.include = set(conf.get('include', []))
        self.exclude = set(conf.get('exclude', []))

    def add_app_handler(self, handler):
        self.app_handlers.append(handler)

    def call_app_handlers(self, msgs):
        for handler in self.app_handlers:
            handler(msgs)

    def handler(self, events_map=None):
        self.msgs = []
        self.run(events_map)
        if self.msgs:
            self.call_app_handlers(self.msgs)

    def include_exclude_prune_set(self, pset):
        if self.include:
            return pset & self.include
        elif self.exclude:
            return pset - self.exclude
        return pset

    def include_exclude_prune(self, lst, entries=2):
        send_lst = []
        for item in lst:
            if item[1] == '-' or item[1].lower().startswith('rotten'):
                continue
            if entries == 1:
                if self.include:
                    if item[0] in self.include:
                        send_lst.append(item[0])
                elif self.exclude:
                    if item[0] not in self.exclude:
                        send_lst.append(item[0])
                else:
                    send_lst.append(item[0])
            else:
                host = item[0]
                host_peer = '%s %s' % (item[0], item[1])

                if self.include:
                    if host in self.include or host_peer in self.include:
                        send_lst.append(host_peer)
                elif self.exclude:
                    if ((host not in self.exclude) and
                            (host_peer not in self.exclude)):
                        send_lst.append(host_peer)
                else:
                    send_lst.append(host_peer)
        # get unique sorted entries
        return sorted(set(send_lst))

    def msg_session_up(self):
        self.msgs.append(('ok', self.log_str + ': All sessions are up'))

    def msg_session_down(self, cnt, lst):
        self.msgs.append(
            ('warning', (self.log_str + ': %d session(s) are down') % cnt, lst))

    def msg_sensor_up(self):
        self.msgs.append(('ok', self.log_str + ': All sensors are ok'))

    def msg_sensor_down(self, cnt, lst):
        self.msgs.append(
            ('warning',
             (self.log_str + ': %d sensor(s) have failures') % cnt, lst))

    def msg_license_up(self):
        self.msgs.append(('ok', self.log_str + ': All licenses are valid'))

    def msg_link_up(self, cnt, lst):
        self.msgs.append(('ok', self.log_str + ': %d link(s) are up' % cnt,
                          lst))

    def msg_link_down(self, cnt, lst):
        self.msgs.append(('warning',
                          self.log_str + ': %d link(s) are down' % cnt, lst))

    def msg_link_flap_up(self, cnt, lst):
        self.msgs.append(('warning', self.log_str +
                          ': %d link(s) flapped and are up' % cnt, lst))

    def msg_link_flap_down(self, cnt, lst):
        self.msgs.append(('warning', self.log_str +
                          ': %d link(s) flapped and are down' % cnt, lst))

    def msg_license_down(self, cnt, lst):
        self.msgs.append(
            ('warning', (self.log_str + ': %d license(s) have failures') % cnt,
             lst))

    def msg_node_ok(self):
        self.msgs.append(('ok', self.log_str + ': All nodes are up'))

    def msg_node_fail(self, cnt, lst):
        self.msgs.append(
            ('warning', (self.log_str + ': %d node(s) have failures') % cnt, lst))

    def msg_node_rotten(self, cnt, lst):
        self.msgs.append(
            ('warning', (self.log_str + ': %d node(s) are rotten') % cnt, lst))

    def msg_mismatch_no(self):
        self.msgs.append(('ok', self.log_str + ': No mismatch found'))

    def msg_mismatch(self, cnt, lst):
        self.msgs.append(
            ('warning', (self.log_str + ': %d mismatch(es) found') % cnt, lst))

    def session_run(self, session_func, msg_down, msg_up):
        # ospf has a different format
        if self.__class__.__name__ == 'Ospf':
            failed_out, counts = session_func()
        else:
            _, counts, failed_out = session_func()
        failed_set = self.include_exclude_prune(failed_out)
        if failed_set != self.failed_set:
            cnt = len(failed_set)
            self.failed_set = failed_set
            if cnt:
                msg_down(cnt, failed_set)
            else:
                msg_up()

    def node_run(self, node_func, warn=False):
        warn_out = []
        if warn:
            _, warn_out, out, _ = node_func()
        else:
            _, out, _ = node_func()
        failed_set = self.include_exclude_prune(warn_out + out, entries=1)
        if failed_set != self.failed_set:
            cnt = len(failed_set)
            self.failed_set = failed_set
            if cnt:
                self.msg_node_fail(cnt, failed_set)
            else:
                self.msg_node_ok()

    def mismatch_run(self, node_func):
        _, _, output = node_func()
        failed_set = self.include_exclude_prune(output)
        if failed_set != self.failed_set:
            cnt = len(failed_set)
            self.failed_set = failed_set
            if cnt:
                self.msg_mismatch(cnt, failed_set)
            else:
                self.msg_mismatch_no()

    def rotten_run(self):
        failed_set = self.include_exclude_prune_set(
                        set(self.lib_netq.rotten_nodes))
        if self.failed_set != failed_set:
            self.failed_set = failed_set
            cnt = len(self.failed_set)
            if cnt:
                self.msg_node_rotten(cnt, failed_set)
            else:
                self.msg_node_ok()


class Bgp(Notifier):
    trigger_events = ['BgpSession', ]

    def run(self, events_map):
        return self.session_run(self.lib_netq.check_bgp_status,
                                self.msg_session_down, self.msg_session_up)


class Ospf(Notifier):
    trigger_events = ['OspfNbr', 'OspfIf']

    def run(self, events_map):
        return self.session_run(self.lib_netq.check_ospf_status,
                                self.msg_session_down, self.msg_session_up)


class Sensors(Notifier):
    trigger_events = ['PSU', 'Fan', 'Temp']

    def run(self, events_map):
        return self.session_run(self.lib_netq.check_sensors,
                                self.msg_sensor_down, self.msg_sensor_up)


class License(Notifier):
    trigger_events = ['License']

    def run(self, events_map):
        return self.session_run(self.lib_netq.check_license,
                                self.msg_license_down, self.msg_license_up)


class Clag(Notifier):
    trigger_events = ['ClagSession', ]

    def run(self, events_map):
        self.node_run(self.lib_netq.check_clag_status, warn=True)


class Lnv(Notifier):
    trigger_events = ['LnvSession', ]

    def run(self, events_map):
        self.node_run(self.lib_netq.check_lnv_status)


class Vxlan(Notifier):
    trigger_events = ['VxlanRemoteDest', 'LnvSession', 'Link']

    def run(self, events_map):
        self.node_run(self.lib_netq.check_vxlan_status, warn=True)


class Vlan(Notifier):
    trigger_events = ['Link']

    def run(self, event_map):
        self.mismatch_run(self.lib_netq.check_vlan_consistency)


class Mtu(Notifier):
    trigger_events = ['Link']

    def run(self, events_map):
        self.mismatch_run(self.lib_netq.check_mtu_consistency)


class Agents(Notifier):

    def run(self, events_map):
        self.rotten_run()


class Link(Notifier):
    trigger_events = ['Link']

    def include_exclude_prune(self, obj):
        ret = None
        item = '%s %s' % (obj.hostname, obj.ifname)
        if self.include:
            if item in self.include or obj.hostname in self.include:
                ret = obj
        elif self.exclude:
            if item not in self.exclude and obj.hostname not in self.exclude:
                ret = obj
        else:
            ret = obj
        return ret

    def run(self, events_map):
        if not events_map:
            return
        up = set()
        down = set()
        flap_down = set()
        flap_up = set()
        for key, data in events_map.items():
            obj = models.Link.key_to_obj(key)
            link_name = obj.hostname + ' ' + obj.ifname
            if not self.include_exclude_prune(obj):
                continue
            cnt = data[1]
            first = True
            flap = 0
            last_st = None
            for link in models.Link.query.range(
                    reverse=True, start=0, end=cnt, first=False,
                    hostname=obj.hostname, ifname=obj.ifname, kind=obj.kind,
                    vni=obj.vni, master=obj.master):
                if first:
                    last_st = link.oper_state
                    curr_st = last_st
                    first = False
                else:
                    if link.oper_state != last_st:
                        last_st = link.oper_state
                        flap += 1

            if flap:
                if flap > 1 and curr_st == 'up':
                    flap_up.add(link_name)
                elif flap > 1 and curr_st == 'down':
                    flap_down.add(link_name)
                elif curr_st == 'down':
                    down.add(link_name)
                elif curr_st == 'up':
                    up.add(link_name)

        if flap_up:
            self.msg_link_flap_up(len(flap_up), flap_up)
        if flap_down:
            self.msg_link_flap_down(len(flap_down), flap_down)
        if up:
            self.msg_link_up(len(up), up)
        if down:
            self.msg_link_down(len(down), down)
